@@ -5,12 +5,12 @@ import dill as pickle
 import jax.numpy as np
 import numpy as onp
 import pandas as pd
-from jax import grad, jit, ops, random, vmap
+from jax import grad, jit, ops, random
 from jax.experimental.optimizers import Optimizer, adam
 
 from pzflow import distributions
 from pzflow.bijectors import Bijector_Info, InitFunction, Pytree
-from pzflow.utils import build_bijector_from_info
+from pzflow.utils import build_bijector_from_info, gaussian_error_model
 
 
 class Flow:
@@ -36,6 +36,9 @@ class Flow:
         bijector: Tuple[InitFunction, Bijector_Info] = None,
         conditional_columns: Sequence[str] = None,
         latent=None,
+        data_error_model: Callable = None,
+        condition_error_model: Callable = None,
+        autoscale_conditions: bool = True,
         seed: int = 0,
         info: Any = None,
         file: str = None,
@@ -63,6 +66,31 @@ class Flow:
             the distributions from pzflow.distributions. If not provided,
             a normal distribution is used with the number of dimensions
             inferred.
+        data_error_model : Callable, optional
+            A callable that defines the error model for data variables.
+            data_error_model must take key, X, Xerr, nsamples as arguments where:
+                key is a jax rng key, e.g. jax.random.PRNGKey(0)
+                X is a 2 dimensional array of data variables, where the order
+                    of variables matches the order of the columns in data_columns
+                Xerr is the corresponding 2 dimensional array of errors
+                nsamples is the number of samples to draw from the error distribution
+            data_error_model must return an array of samples with the shape
+            (X.shape[0], nsamples, X.shape[1]).
+            If data_error_model is not provided, a Gaussian error model is assumed.
+        condition_error_model : Callable, optional
+            A callable that defines the error model for conditional variables.
+            condition_error_model must take key, X, Xerr, nsamples as arguments where:
+                key is a jax rng key, e.g. jax.random.PRNGKey(0)
+                X is a 2 dimensional array of conditional variables, where the order
+                    of variables matches the order of the columns in conditional_columns
+                Xerr is the corresponding 2 dimensional array of errors
+                nsamples is the number of samples to draw from the error distribution
+            condition_error_model must return an array of samples with the shape
+            (X.shape[0], nsamples, X.shape[1]).
+            If condition_error_model is not provided, a Gaussian error model is assumed.
+        autoscale_conditions : bool, default=True
+            Sets whether or not conditions are automatically standard scaled when
+            passed to a conditional flow. I recommend you leave this as True.
         seed : int, default=0
             The random seed for initial parameters
         info : Any, optional
@@ -90,6 +118,8 @@ class Flow:
                 bijector is not None,
                 conditional_columns is not None,
                 latent is not None,
+                data_error_model is not None,
+                condition_error_model is not None,
                 info is not None,
             )
         ):
@@ -132,6 +162,10 @@ class Flow:
                 *self._latent_info[1]
             )
 
+            # load the error models
+            self.data_error_model = save_dict["data_error_model"]
+            self.condition_error_model = save_dict["condition_error_model"]
+
             # load the bijector
             self._bijector_info = save_dict["bijector_info"]
             init_fun, _ = build_bijector_from_info(self._bijector_info)
@@ -144,6 +178,10 @@ class Flow:
             self._condition_means = save_dict["condition_means"]
             self._condition_stds = save_dict["condition_stds"]
 
+            # set whether or not to automatically standard scale any
+            # conditions passed to the normalizing flow
+            self._autoscale_conditions = save_dict["autoscale_conditions"]
+
         # if no file is provided, use provided parameters
         else:
             self.data_columns = tuple(data_columns)
@@ -159,6 +197,10 @@ class Flow:
                 self._condition_means = np.zeros(len(self.conditional_columns))
                 self._condition_stds = np.ones(len(self.conditional_columns))
 
+            # set whether or not to automatically standard scale any
+            # conditions passed to the normalizing flow
+            self._autoscale_conditions = autoscale_conditions
+
             # set up the latent distribution
             if latent is None:
                 self.latent = distributions.Normal(self._input_dim)
@@ -166,6 +208,16 @@ class Flow:
                 self.latent = latent
             self._latent_info = self.latent.info
 
+            # set up the error models
+            if data_error_model is None:
+                self.data_error_model = gaussian_error_model
+            else:
+                self.data_error_model = data_error_model
+            if condition_error_model is None:
+                self.condition_error_model = gaussian_error_model
+            else:
+                self.condition_error_model = condition_error_model
+
             # set up the bijector with random params
             init_fun, self._bijector_info = bijector
             bijector_params, self._forward, self._inverse = init_fun(
@@ -186,26 +238,28 @@ class Flow:
             conditions = (conditions - self._condition_means) / self._condition_stds
         return conditions
 
-    def _get_samples(
+    def _get_err_samples(
         self,
-        rng,
+        key,
         inputs: pd.DataFrame,
-        nsamples: int,
+        err_samples: int,
         type: str = "data",
         skip: str = None,
     ) -> np.ndarray:
-        """Draw Gaussian samples for each row of inputs. """
+        """Draw error samples for each row of inputs. """
 
         X = inputs.copy()
 
         # get list of columns
         if type == "data":
             columns = list(self.data_columns)
+            error_model = self.data_error_model
         elif type == "conditions":
             if self.conditional_columns is None:
-                return np.zeros((nsamples * X.shape[0], 1))
+                return np.zeros((err_samples * X.shape[0], 1))
             else:
                 columns = list(self.conditional_columns)
+                error_model = self.condition_error_model
         else:
             raise ValueError("type must be `data` or `conditions`.")
 
@@ -214,20 +268,23 @@ class Flow:
             # if errors not provided for the column, fill in zeros
             if f"{col}_err" not in inputs.columns and col != skip:
                 X[f"{col}_err"] = np.zeros(X.shape[0])
+            # if we are skipping this column, fill in nan's
+            elif col == skip:
+                X[col] = np.nan * np.zeros(X.shape[0])
+                X[f"{col}_err"] = np.nan * np.zeros(X.shape[0])
 
         # pull out relevant columns
-        if skip is not None:
-            columns.remove(skip)
         err_columns = [col + "_err" for col in columns]
         X, Xerr = np.array(X[columns].values), np.array(X[err_columns].values)
 
-        # lower bound on Xerr to avoid singular matrix
-        Xerr = np.clip(Xerr, 1e-8, None)
         # generate samples
-        Xsamples = random.multivariate_normal(
-            rng, X, vmap(np.diag)(Xerr ** 2), shape=(nsamples, X.shape[0])
-        )
-        Xsamples = Xsamples.reshape(-1, X.shape[1], order="F")
+        Xsamples = error_model(key, X, Xerr, err_samples)
+        Xsamples = Xsamples.reshape(X.shape[0] * err_samples, X.shape[1])
+
+        # delete the column corresponding to skip
+        if skip is not None:
+            idx = columns.index(skip)
+            Xsamples = np.delete(Xsamples, idx, axis=1)
 
         # if these are samples of conditions, standard scale them!
         if type == "conditions":
@@ -246,32 +303,8 @@ class Flow:
         log_prob = np.nan_to_num(log_prob, nan=np.NINF)
         return log_prob
 
-    def _save_dict(self):
-        """Returns the dictionary of all flow params to be saved."""
-        save_dict = {"class": "Flow"}
-        keys = [
-            "data_columns",
-            "conditional_columns",
-            "condition_means",
-            "condition_stds",
-            "info",
-            "latent_info",
-            "bijector_info",
-            "params",
-        ]
-        for key in keys:
-            try:
-                save_dict[key] = getattr(self, key)
-            except AttributeError:
-                try:
-                    save_dict[key] = getattr(self, "_" + key)
-                except AttributeError:
-                    save_dict[key] = None
-
-        return save_dict
-
     def log_prob(
-        self, inputs: pd.DataFrame, nsamples: int = None, seed: int = None
+        self, inputs: pd.DataFrame, err_samples: int = None, seed: int = None
     ) -> np.ndarray:
         """Calculates log probability density of inputs.
 
@@ -282,12 +315,12 @@ class Flow:
             Every column in self.data_columns must be present.
             If self.conditional_columns is not None, those must be present
             as well. If other columns are present, they are ignored.
-        nsamples : int, default=None
-            Number of samples to average over for the log_prob calculation.
-            If provided, then Gaussian errors are assumed, and method will
-            look for error columns in `inputs`. Error columns must end in
-            `_err`. E.g. the error column for the variable `u` must be `u_err`.
-            Zero error assumed for any missing error columns.
+        err_samples : int, default=None
+            Number of samples from the error distribution to average over for
+            the log_prob calculation. If provided, Gaussian errors are assumed,
+            and method will look for error columns in `inputs`. Error columns
+            must end in `_err`. E.g. the error column for the variable `u` must
+            be `u_err`. Zero error assumed for any missing error columns.
         seed : int, default=None
             Random seed for drawing the samples with Gaussian errors.
 
@@ -297,7 +330,7 @@ class Flow:
             Device array of shape (inputs.shape[0],).
         """
 
-        if nsamples is None:
+        if err_samples is None:
             # convert data to an array with columns ordered
             columns = list(self.data_columns)
             X = np.array(inputs[columns].values)
@@ -308,16 +341,18 @@ class Flow:
 
         else:
             # validate nsamples
-            assert isinstance(nsamples, int), "nsamples must be a positive integer."
-            assert nsamples > 0, "nsamples must be a positive integer."
+            assert isinstance(
+                err_samples, int
+            ), "err_samples must be a positive integer."
+            assert err_samples > 0, "err_samples must be a positive integer."
             # get Gaussian samples
             seed = onp.random.randint(1e18) if seed is None else seed
-            rng = random.PRNGKey(seed)
-            X = self._get_samples(rng, inputs, nsamples, type="data")
-            C = self._get_samples(rng, inputs, nsamples, type="conditions")
+            key = random.PRNGKey(seed)
+            X = self._get_err_samples(key, inputs, err_samples, type="data")
+            C = self._get_err_samples(key, inputs, err_samples, type="conditions")
             # calculate log_probs
             log_probs = self._log_prob(self._params, X, C)
-            probs = np.exp(log_probs.reshape(-1, nsamples))
+            probs = np.exp(log_probs.reshape(-1, err_samples))
             return np.log(probs.mean(axis=1))
 
     def posterior(
@@ -325,10 +360,12 @@ class Flow:
         inputs: pd.DataFrame,
         column: str,
         grid: np.ndarray,
+        marg_rules: dict = None,
         normalize: bool = True,
-        nsamples: int = None,
+        err_samples: int = None,
         seed: int = None,
         batch_size: int = None,
+        nan_to_zero: bool = True,
     ) -> np.ndarray:
         """Calculates posterior distributions for the provided column.
 
@@ -348,20 +385,34 @@ class Flow:
             `inputs` is irrelevant.
         grid : np.ndarray
             Grid on which to calculate the posterior.
-        normalize : boolean, default=True
-            Whether to normalize the posterior so that it integrates to 1.
-        nsamples : int, default=None
-            Number of samples to average over for the posterior calculation.
-            If provided, then Gaussian errors are assumed, and method will
-            look for error columns in `inputs`. Error columns must end in
-            `_err`. E.g. the error column for the variable `u` must be `u_err`.
-            Zero error assumed for any missing error columns.
+        marg_rules : dict, optional
+            Dictionary with rules for marginalizing over missing variables.
+            The dictionary must contain the key "flag", which gives the flag
+            that indicates a missing value. E.g. if missing values are given
+            the value 99, the dictionary should contain {"flag": 99}.
+            The dictionary must also contain {"name": callable} for any
+            variables that will need to be marginalized over, where name is
+            the name of the variable, and callable is a callable that takes
+            the row of variables nad returns a grid over which to marginalize
+            the variable. E.g. {"y": lambda row: np.linspace(0, row["x"], 10)}.
+            Note: the callable for a given name must *always* return an array
+            of the same length, regardless of the input row.
+        err_samples : int, default=None
+            Number of samples from the error distribution to average over for
+            the posterior calculation. If provided, Gaussian errors are assumed,
+            and method will look for error columns in `inputs`. Error columns
+            must end in `_err`. E.g. the error column for the variable `u` must
+            be `u_err`. Zero error assumed for any missing error columns.
         seed : int, default=None
             Random seed for drawing the samples with Gaussian errors.
         batch_size : int, default=None
             Size of batches in which to calculate posteriors. If None, all
             posteriors are calculated simultaneously. Simultaneous calculation
             is faster, but memory intensive for large data sets.
+        normalize : boolean, default=True
+            Whether to normalize the posterior so that it integrates to 1.
+        nan_to_zero : bool, default=True
+            Whether to convert NaN's to zero probability in the final pdfs.
 
         Returns
         -------
@@ -377,83 +428,198 @@ class Flow:
         nrows = inputs.shape[0]
         batch_size = nrows if batch_size is None else batch_size
 
-        if nsamples is not None:
+        # make sure indices run 0 -> nrows
+        inputs = inputs.reset_index(drop=True)
+
+        if err_samples is not None:
             # validate nsamples
-            assert isinstance(nsamples, int), "nsamples must be a positive integer."
-            assert nsamples > 0, "nsamples must be a positive integer."
+            assert isinstance(
+                err_samples, int
+            ), "err_samples must be a positive integer."
+            assert err_samples > 0, "err_samples must be a positive integer."
             # set the seed
             seed = onp.random.randint(1e18) if seed is None else seed
-            rng = random.PRNGKey(seed)
+            key = random.PRNGKey(seed)
 
         # empty array to hold pdfs
         pdfs = np.zeros((nrows, len(grid)))
 
-        # loop through batches
-        for batch_idx in range(0, nrows, batch_size):
+        # if marginalization rules were passed, we will loop over the rules
+        # and repeatedly call this method
+        if marg_rules is not None:
 
-            # get the data batch
-            # and, if this is a conditional flow, the correpsonding conditions
-            batch = inputs.iloc[batch_idx : batch_idx + batch_size]
+            # if the flag is NaN, we must use np.isnan to check for flags
+            if onp.isnan(marg_rules["flag"]):
 
-            # if not drawing samples, just grab batch and conditions
-            if nsamples is None:
-                conditions = self._get_conditions(batch)
-                batch = np.array(batch[columns].values)
-            # if only drawing condition samples...
-            elif len(self.data_columns) == 1:
-                conditions = self._get_samples(rng, batch, nsamples, type="conditions")
-                batch = np.repeat(batch[columns].values, nsamples, axis=0)
-            # if drawing data and condition samples...
+                def check_flags(data):
+                    return onp.isnan(data)
+
+            # else we use np.isclose to check for flags
             else:
-                conditions = self._get_samples(rng, batch, nsamples, type="conditions")
-                batch = self._get_samples(
-                    rng, batch, nsamples, skip=column, type="data"
-                )
 
-            # make a new copy of each row for each value of the column
-            # for which we are calculating the posterior
-            batch = np.hstack(
-                (
-                    np.repeat(
-                        batch[:, :idx],
-                        len(grid),
-                        axis=0,
-                    ),
-                    np.tile(grid, len(batch))[:, None],
-                    np.repeat(
-                        batch[:, idx:],
-                        len(grid),
-                        axis=0,
-                    ),
-                )
+                def check_flags(data):
+                    return onp.isclose(data, marg_rules["flag"])
+
+            # first calculate pdfs for unflagged rows
+            unflagged_idx = inputs[
+                ~check_flags(inputs[columns]).any(axis=1)
+            ].index.tolist()
+            unflagged_pdfs = self.posterior(
+                inputs=inputs.iloc[unflagged_idx],
+                column=column,
+                grid=grid,
+                err_samples=err_samples,
+                seed=seed,
+                batch_size=batch_size,
+                normalize=False,
+                nan_to_zero=nan_to_zero,
             )
 
-            # make similar copies of the conditions
-            conditions = np.repeat(conditions, len(grid), axis=0)
-
-            # calculate probability densities
-            log_prob = self._log_prob(self._params, batch, conditions).reshape(
-                (-1, len(grid))
-            )
-            prob = np.exp(log_prob)
-            # if we were Gaussian sampling, average over the samples
-            if nsamples is not None:
-                prob = prob.reshape(-1, nsamples, len(grid))
-                prob = prob.mean(axis=1)
-            # add the pdfs to the bigger list
+            # save these pdfs in the big array
             pdfs = ops.index_update(
                 pdfs,
-                ops.index[batch_idx : batch_idx + batch_size, :],
-                prob,
+                ops.index[unflagged_idx, :],
+                unflagged_pdfs,
                 indices_are_sorted=True,
                 unique_indices=True,
             )
 
+            # we will keep track of all the rows we've already calculated
+            # posteriors for
+            already_done = unflagged_idx
+
+            # now we will loop over the rules in marg_rules
+            for name, rule in marg_rules.items():
+
+                # ignore the flag, because that's not a column in the data
+                if name == "flag":
+                    continue
+
+                # get the list of new rows for which we need to calculate posteriors
+                flagged_idx = inputs[check_flags(inputs[name])].index.tolist()
+                flagged_idx = list(set(flagged_idx).difference(already_done))
+
+                # if flagged_idx is empty, move on!
+                if len(flagged_idx) == 0:
+                    continue
+
+                # get the marginalization grid for each row
+                marg_grids = (
+                    inputs.iloc[flagged_idx]
+                    .apply(rule, axis=1, result_type="expand")
+                    .values
+                )
+
+                # make a new data frame with the marginalization grids replacing
+                # the values of the flag in the column
+                marg_inputs = pd.DataFrame(
+                    np.repeat(
+                        inputs.iloc[flagged_idx].values, marg_grids.shape[1], axis=0
+                    ),
+                    columns=inputs.columns,
+                )
+                marg_inputs[name] = marg_grids.reshape(marg_inputs.shape[0], 1)
+
+                # remove the error column if it's present
+                marg_inputs.drop(f"{name}_err", axis=1, inplace=True, errors="ignore")
+
+                # calculate posteriors for these
+                marg_pdfs = self.posterior(
+                    inputs=marg_inputs,
+                    column=column,
+                    grid=grid,
+                    marg_rules=marg_rules,
+                    err_samples=err_samples,
+                    seed=seed,
+                    batch_size=batch_size,
+                    normalize=False,
+                    nan_to_zero=nan_to_zero,
+                )
+
+                # sum over the marginalized dimension
+                marg_pdfs = marg_pdfs.reshape(
+                    len(flagged_idx), marg_grids.shape[1], grid.size
+                )
+                marg_pdfs = marg_pdfs.sum(axis=1)
+
+                # save the new pdfs in the big array
+                pdfs = ops.index_update(
+                    pdfs,
+                    ops.index[flagged_idx, :],
+                    marg_pdfs,
+                    indices_are_sorted=True,
+                    unique_indices=True,
+                )
+
+                # add these flagged indices to the list of rows already done
+                already_done += flagged_idx
+
+        # now for the main posterior calculation loop
+        else:
+
+            # loop through batches
+            for batch_idx in range(0, nrows, batch_size):
+
+                # get the data batch
+                # and, if this is a conditional flow, the correpsonding conditions
+                batch = inputs.iloc[batch_idx : batch_idx + batch_size]
+
+                # if not drawing samples, just grab batch and conditions
+                if err_samples is None:
+                    conditions = self._get_conditions(batch)
+                    batch = np.array(batch[columns].values)
+                # if only drawing condition samples...
+                elif len(self.data_columns) == 1:
+                    conditions = self._get_err_samples(
+                        key, batch, err_samples, type="conditions"
+                    )
+                    batch = np.repeat(batch[columns].values, err_samples, axis=0)
+                # if drawing data and condition samples...
+                else:
+                    conditions = self._get_err_samples(
+                        key, batch, err_samples, type="conditions"
+                    )
+                    batch = self._get_err_samples(
+                        key, batch, err_samples, skip=column, type="data"
+                    )
+
+                # make a new copy of each row for each value of the column
+                # for which we are calculating the posterior
+                batch = np.hstack(
+                    (
+                        np.repeat(batch[:, :idx], len(grid), axis=0,),
+                        np.tile(grid, len(batch))[:, None],
+                        np.repeat(batch[:, idx:], len(grid), axis=0,),
+                    )
+                )
+
+                # make similar copies of the conditions
+                conditions = np.repeat(conditions, len(grid), axis=0)
+
+                # calculate probability densities
+                log_prob = self._log_prob(self._params, batch, conditions).reshape(
+                    (-1, len(grid))
+                )
+                prob = np.exp(log_prob)
+                # if we were Gaussian sampling, average over the samples
+                if err_samples is not None:
+                    prob = prob.reshape(-1, err_samples, len(grid))
+                    prob = prob.mean(axis=1)
+                # add the pdfs to the bigger list
+                pdfs = ops.index_update(
+                    pdfs,
+                    ops.index[batch_idx : batch_idx + batch_size, :],
+                    prob,
+                    indices_are_sorted=True,
+                    unique_indices=True,
+                )
+
         if normalize:
             # normalize so they integrate to one
             pdfs = pdfs / np.trapz(y=pdfs, x=grid).reshape(-1, 1)
-        # set NaN's equal to zero probability
-        pdfs = np.nan_to_num(pdfs, nan=0.0)
+        if nan_to_zero:
+            # set NaN's equal to zero probability
+            pdfs = np.nan_to_num(pdfs, nan=0.0)
         return pdfs
 
     def sample(
@@ -521,6 +687,33 @@ class Flow:
         # return the samples!
         return x
 
+    def _save_dict(self):
+        """Returns the dictionary of all flow params to be saved."""
+        save_dict = {"class": self.__class__.__name__}
+        keys = [
+            "data_columns",
+            "conditional_columns",
+            "condition_means",
+            "condition_stds",
+            "data_error_model",
+            "condition_error_model",
+            "autoscale_conditions",
+            "info",
+            "latent_info",
+            "bijector_info",
+            "params",
+        ]
+        for key in keys:
+            try:
+                save_dict[key] = getattr(self, key)
+            except AttributeError:
+                try:
+                    save_dict[key] = getattr(self, "_" + key)
+                except AttributeError:
+                    save_dict[key] = None
+
+        return save_dict
+
     def save(self, file: str):
         """Saves the flow to a file.
 
@@ -540,8 +733,6 @@ class Flow:
         """
         save_dict = self._save_dict()
 
-        if not file.endswith(".pkl"):
-            file += ".pkl"
         with open(file, "wb") as handle:
             pickle.dump(save_dict, handle, recurse=True)
 
@@ -552,7 +743,7 @@ class Flow:
         batch_size: int = 1024,
         optimizer: Optimizer = None,
         loss_fn: Callable = None,
-        sample_errs: bool = False,
+        convolve_errs: bool = False,
         seed: int = 0,
         verbose: bool = False,
     ) -> list:
@@ -572,7 +763,7 @@ class Flow:
         loss_fn : Callable, optional
             A function to calculate the loss: loss = loss_fn(params, x).
             If not provided, will be -mean(log_prob).
-        sample_errs : bool, default=False
+        convolve_errs : bool, default=False
             Whether to draw new data from the error distributions during
             each epoch of training. Assumes errors are Gaussian, and method
             will look for error columns in `inputs`. Error columns must end
@@ -616,31 +807,33 @@ class Flow:
         # get list of data columns
         columns = list(self.data_columns)
 
-        # if this is a conditional flow, save the means and stds of the conditional columns
-        if self.conditional_columns is not None:
+        # if this is a conditional flow, and autoscale_conditions == True
+        # save the means and stds of the conditional columns
+        if self.conditional_columns is not None and self._autoscale_conditions:
             self._condition_means = np.array(
                 inputs[list(self.conditional_columns)].values.mean(axis=0)
             )
-            self._condition_stds = np.array(
+            condition_stds = np.array(
                 inputs[list(self.conditional_columns)].values.std(axis=0)
             )
+            self._condition_stds = np.where(condition_stds != 0, condition_stds, 1)
 
         # define a function to return batches
-        if sample_errs:
+        if convolve_errs:
 
-            def get_batch(sample_rng, x, type):
-                return self._get_samples(sample_rng, x, 1, type=type)
+            def get_batch(sample_key, x, type):
+                return self._get_err_samples(sample_key, x, 1, type=type)
 
         else:
 
-            def get_batch(sample_rng, x, type):
+            def get_batch(sample_key, x, type):
                 if type == "conditions":
                     return self._get_conditions(x)
                 else:
                     return np.array(x[columns].values)
 
         # get random seed for training loop
-        rng = random.PRNGKey(seed)
+        key = random.PRNGKey(seed)
 
         if verbose:
             print(f"Training {epochs} epochs \nLoss:")
@@ -656,8 +849,8 @@ class Flow:
         itercount = itertools.count()
         for epoch in range(epochs):
             # new permutation of batches
-            permute_rng, sample_rng, rng = random.split(rng, num=3)
-            idx = random.permutation(permute_rng, inputs.shape[0])
+            permute_key, sample_key, key = random.split(key, num=3)
+            idx = random.permutation(permute_key, inputs.shape[0])
             X = inputs.iloc[idx]
 
             # loop through batches and step optimizer
@@ -667,29 +860,20 @@ class Flow:
                 # Gaussian sample of the batch. Else just returns batch as a
                 # jax array
                 batch = get_batch(
-                    sample_rng, X.iloc[batch_idx : batch_idx + batch_size], type="data"
+                    sample_key, X.iloc[batch_idx : batch_idx + batch_size], type="data"
                 )
                 batch_conditions = get_batch(
-                    sample_rng,
+                    sample_key,
                     X.iloc[batch_idx : batch_idx + batch_size],
                     type="conditions",
                 )
 
-                opt_state = step(
-                    next(itercount),
-                    opt_state,
-                    batch,
-                    batch_conditions,
-                )
+                opt_state = step(next(itercount), opt_state, batch, batch_conditions,)
 
             # save end-of-epoch training loss
             params = get_params(opt_state)
             losses.append(
-                loss_fn(
-                    params,
-                    np.array(X[columns].values),
-                    self._get_conditions(X),
-                )
+                loss_fn(params, np.array(X[columns].values), self._get_conditions(X),)
             )
 
             if verbose and (
