@@ -2,7 +2,9 @@ from typing import Any, Callable, Sequence, Tuple
 
 import dill as pickle
 import jax.numpy as np
+import numpy as onp
 import pandas as pd
+from jax import random
 from jax.experimental.optimizers import Optimizer
 
 from pzflow import Flow
@@ -32,6 +34,9 @@ class FlowEnsemble:
         bijector: Tuple[InitFunction, Bijector_Info] = None,
         conditional_columns: Sequence[str] = None,
         latent=None,
+        data_error_model: Callable = None,
+        condition_error_model: Callable = None,
+        autoscale_conditions: bool = True,
         N: int = 1,
         info: Any = None,
         file: str = None,
@@ -58,6 +63,31 @@ class FlowEnsemble:
             the distributions from pzflow.distributions. If not provided,
             a normal distribution is used with the number of dimensions
             inferred.
+        data_error_model : Callable, optional
+            A callable that defines the error model for data variables.
+            data_error_model must take key, X, Xerr, nsamples as arguments where:
+                key is a jax rng key, e.g. jax.random.PRNGKey(0)
+                X is a 2 dimensional array of data variables, where the order
+                    of variables matches the order of the columns in data_columns
+                Xerr is the corresponding 2 dimensional array of errors
+                nsamples is the number of samples to draw from the error distribution
+            data_error_model must return an array of samples with the shape
+            (X.shape[0], nsamples, X.shape[1]).
+            If data_error_model is not provided, a Gaussian error model is assumed.
+        condition_error_model : Callable, optional
+            A callable that defines the error model for conditional variables.
+            condition_error_model must take key, X, Xerr, nsamples as arguments where:
+                key is a jax rng key, e.g. jax.random.PRNGKey(0)
+                X is a 2 dimensional array of conditional variables, where the order
+                    of variables matches the order of the columns in conditional_columns
+                Xerr is the corresponding 2 dimensional array of errors
+                nsamples is the number of samples to draw from the error distribution
+            condition_error_model must return an array of samples with the shape
+            (X.shape[0], nsamples, X.shape[1]).
+            If condition_error_model is not provided, a Gaussian error model is assumed.
+        autoscale_conditions : bool, defautl=True
+            Sets whether or not conditions are automatically standard scaled when
+            passed to a conditional flow. I recommend you leave this as True.
         N : int, default=1
             The number of flows in the ensemble.
         info : Any, optional
@@ -80,6 +110,8 @@ class FlowEnsemble:
                 bijector is not None,
                 conditional_columns is not None,
                 latent is not None,
+                data_error_model is not None,
+                condition_error_model is not None,
                 info is not None,
             )
         ):
@@ -104,25 +136,39 @@ class FlowEnsemble:
             # load the ensemble from the dictionary
             self._ensemble = {
                 name: Flow(_dictionary=flow_dict)
-                for name, flow_dict in save_dict.items()
+                for name, flow_dict in save_dict["ensemble"].items()
             }
+            # load the metadata
+            self.data_columns = save_dict["data_columns"]
+            self.conditional_columns = save_dict["conditional_columns"]
+            self.info = save_dict["info"]
+
         # otherwise create a new ensemble from the provided parameters
         else:
+            # save the ensemble of flows
             self._ensemble = {
                 f"Flow {i}": Flow(
                     data_columns=data_columns,
                     bijector=bijector,
                     conditional_columns=conditional_columns,
                     latent=latent,
+                    data_error_model=data_error_model,
+                    condition_error_model=condition_error_model,
+                    autoscale_conditions=autoscale_conditions,
                     seed=i,
+                    info=f"Flow {i}",
                 )
                 for i in range(N)
             }
+            # save the metadata
+            self.data_columns = data_columns
+            self.conditional_columns = conditional_columns
+            self.info = info
 
     def log_prob(
         self,
         inputs: pd.DataFrame,
-        nsamples: int = None,
+        err_samples: int = None,
         seed: int = None,
         returnEnsemble: bool = False,
     ) -> np.ndarray:
@@ -135,12 +181,12 @@ class FlowEnsemble:
             Every column in self.data_columns must be present.
             If self.conditional_columns is not None, those must be present
             as well. If other columns are present, they are ignored.
-        nsamples : int, default=None
-            Number of samples to average over for the log_prob calculation.
-            If provided, then Gaussian errors are assumed, and method will
-            look for error columns in `inputs`. Error columns must end in
-            `_err`. E.g. the error column for the variable `u` must be `u_err`.
-            Zero error assumed for any missing error columns.
+        err_samples : int, default=None
+            Number of samples from the error distribution to average over for
+            the log_prob calculation. If provided, Gaussian errors are assumed,
+            and method will look for error columns in `inputs`. Error columns
+            must end in `_err`. E.g. the error column for the variable `u` must
+            be `u_err`. Zero error assumed for any missing error columns.
         seed : int, default=None
             Random seed for drawing the samples with Gaussian errors.
         returnEnsemble : bool, default=False
@@ -158,7 +204,10 @@ class FlowEnsemble:
 
         # calculate log_prob for each flow in the ensemble
         ensemble = np.array(
-            [flow.log_prob(inputs, nsamples, seed) for flow in self._ensemble.values()]
+            [
+                flow.log_prob(inputs, err_samples, seed)
+                for flow in self._ensemble.values()
+            ]
         )
 
         # re-arrange so that (axis 0, axis 1) = (inputs, flows in ensemble)
@@ -177,11 +226,13 @@ class FlowEnsemble:
         inputs: pd.DataFrame,
         column: str,
         grid: np.ndarray,
+        marg_rules: dict = None,
         normalize: bool = True,
-        nsamples: int = None,
+        err_samples: int = None,
         seed: int = None,
         batch_size: int = None,
         returnEnsemble: bool = False,
+        nan_to_zero: bool = True,
     ) -> np.ndarray:
         """Calculates posterior distributions for the provided column.
 
@@ -201,14 +252,26 @@ class FlowEnsemble:
             `inputs` is irrelevant.
         grid : np.ndarray
             Grid on which to calculate the posterior.
+        marg_rules : dict, optional
+            Dictionary with rules for marginalizing over missing variables.
+            The dictionary must contain the key "flag", which gives the flag
+            that indicates a missing value. E.g. if missing values are given
+            the value 99, the dictionary should contain {"flag": 99}.
+            The dictionary must also contain {"name": callable} for any
+            variables that will need to be marginalized over, where name is
+            the name of the variable, and callable is a callable that takes
+            the row of variables nad returns a grid over which to marginalize
+            the variable. E.g. {"y": lambda row: np.linspace(0, row["x"], 10)}.
+            Note: the callable for a given name must *always* return an array
+            of the same length, regardless of the input row.
         normalize : boolean, default=True
             Whether to normalize the posterior so that it integrates to 1.
-        nsamples : int, default=None
-            Number of samples to average over for the posterior calculation.
-            If provided, then Gaussian errors are assumed, and method will
-            look for error columns in `inputs`. Error columns must end in
-            `_err`. E.g. the error column for the variable `u` must be `u_err`.
-            Zero error assumed for any missing error columns.
+        err_samples : int, default=None
+            Number of samples from the error distribution to average over for
+            the posterior calculation. If provided, Gaussian errors are assumed,
+            and method will look for error columns in `inputs`. Error columns
+            must end in `_err`. E.g. the error column for the variable `u` must
+            be `u_err`. Zero error assumed for any missing error columns.
         seed : int, default=None
             Random seed for drawing the samples with Gaussian errors.
         batch_size : int, default=None
@@ -220,6 +283,8 @@ class FlowEnsemble:
             array of shape (inputs.shape[0], N flows in ensemble, grid.size).
             If False, the posterior is averaged over the flows in the ensemble,
             and returned as an array of shape (inputs.shape[0], grid.size)
+        nan_to_zero : bool, default=True
+            Whether to convert NaN's to zero probability in the final pdfs.
 
         Returns
         -------
@@ -230,7 +295,17 @@ class FlowEnsemble:
         # calculate posterior for each flow in the ensemble
         ensemble = np.array(
             [
-                flow.posterior(inputs, column, grid, False, nsamples, seed, batch_size)
+                flow.posterior(
+                    inputs=inputs,
+                    column=column,
+                    grid=grid,
+                    marg_rules=marg_rules,
+                    err_samples=err_samples,
+                    seed=seed,
+                    batch_size=batch_size,
+                    normalize=False,
+                    nan_to_zero=nan_to_zero,
+                )
                 for flow in self._ensemble.values()
             ]
         )
@@ -260,7 +335,9 @@ class FlowEnsemble:
         seed: int = None,
         returnEnsemble: bool = False,
     ) -> pd.DataFrame:
-        """Returns samples from the normalizing flow.
+        """Returns samples from the ensemble.
+
+        # UPDATE THIS DOCSTRING TO REFLECT THE NEW METHOD FOR CONDITIONAL SAMPLING
 
         Parameters
         ----------
@@ -295,15 +372,82 @@ class FlowEnsemble:
                 keys=self._ensemble.keys(),
             )
         else:
-            # return nsamples drawn uniformly from the flows in the ensemble
-            N = int(np.ceil(nsamples / len(self._ensemble)))
-            samples = pd.concat(
-                [
-                    flow.sample(N, conditions, save_conditions, seed)
-                    for flow in self._ensemble.values()
-                ]
-            )
-            return samples.sample(nsamples, random_state=seed).reset_index(drop=True)
+            # if this isn't a conditional flow, sampling is straightforward
+            if conditions is None:
+                # return nsamples drawn uniformly from the flows in the ensemble
+                N = int(np.ceil(nsamples / len(self._ensemble)))
+                samples = pd.concat(
+                    [
+                        flow.sample(N, conditions, save_conditions, seed)
+                        for flow in self._ensemble.values()
+                    ]
+                )
+                return samples.sample(nsamples, random_state=seed).reset_index(
+                    drop=True
+                )
+            # if this is a conditional flow, it's a little more complicated...
+            else:
+                # if nsamples > 1, we duplicate the rows of the conditions,
+                # and then call the sample method again with nsamples = 1
+                if nsamples > 1:
+                    new_conditions = pd.concat([conditions] * nsamples)
+                    return self.sample(
+                        nsamples=1,
+                        conditions=new_conditions,
+                        save_conditions=save_conditions,
+                        seed=seed,
+                    )
+                # if nsamples = 1, we can proceed with the main sampling algorithm
+                else:
+                    seed = onp.random.randint(1e18) if seed is None else seed
+                    # if we are drawing more samples than the number of flows in
+                    # the ensemble, then we will shuffle the conditions and randomly
+                    # assign them to one of the constituent flows
+                    if conditions.shape[0] > len(self._ensemble):
+                        # shuffle the conditions
+                        conditions_shuffled = conditions.sample(
+                            frac=1.0, random_state=int(seed / 1e9)
+                        )
+                        # split conditions into ~equal sized chunks
+                        chunks = onp.array_split(
+                            conditions_shuffled, len(self._ensemble)
+                        )
+                        # shuffle the chunks
+                        chunks = [
+                            chunks[i]
+                            for i in random.permutation(
+                                random.PRNGKey(seed), np.arange(len(chunks))
+                            )
+                        ]
+                        # sample from each flow, and return all the samples
+                        return pd.concat(
+                            [
+                                flow.sample(1, chunk, save_conditions, seed)
+                                for flow, chunk in zip(self._ensemble.values(), chunks)
+                            ],
+                            ignore_index=True,
+                        )
+                    # however, if there are more flows in the ensemble than samples
+                    # being drawn, then we will randomly select flows for each condition
+                    else:
+                        rng = onp.random.default_rng(seed)
+                        # randomly select a flow to sample from for each condition
+                        flows = rng.choice(
+                            list(self._ensemble.values()),
+                            size=conditions.shape[0],
+                            replace=True,
+                        )
+                        # sample from each flow and return all the samples together
+                        seeds = rng.integers(1e18, size=conditions.shape[0])
+                        return pd.concat(
+                            [
+                                flow.sample(
+                                    1, conditions[i : i + 1], save_conditions, new_seed
+                                )
+                                for i, (flow, new_seed) in enumerate(zip(flows, seeds))
+                            ],
+                            ignore_index=True,
+                        )
 
     def save(self, file: str):
         """Saves the ensemble to a file.
@@ -322,11 +466,16 @@ class FlowEnsemble:
             Path to where the ensemble will be saved.
             Extension `.pkl` will be appended if not already present.
         """
-        save_dict = {name: flow._save_dict() for name, flow in self._ensemble.items()}
-        save_dict["class"] = "FlowEnsemble"
+        save_dict = {
+            "data_columns": self.data_columns,
+            "conditional_columns": self.conditional_columns,
+            "info": self.info,
+            "class": self.__class__.__name__,
+            "ensemble": {
+                name: flow._save_dict() for name, flow in self._ensemble.items()
+            },
+        }
 
-        if not file.endswith(".pkl"):
-            file += ".pkl"
         with open(file, "wb") as handle:
             pickle.dump(save_dict, handle, recurse=True)
 
@@ -337,7 +486,7 @@ class FlowEnsemble:
         batch_size: int = 1024,
         optimizer: Optimizer = None,
         loss_fn: Callable = None,
-        sample_errs: bool = False,
+        convolve_errs: bool = False,
         seed: int = 0,
         verbose: bool = False,
     ) -> dict:
@@ -357,7 +506,7 @@ class FlowEnsemble:
         loss_fn : Callable, optional
             A function to calculate the loss: loss = loss_fn(params, x).
             If not provided, will be -mean(log_prob).
-        sample_errs : bool, default=False
+        convolve_errs : bool, default=False
             Whether to draw new data from the error distributions during
             each epoch of training. Assumes errors are Gaussian, and method
             will look for error columns in `inputs`. Error columns must end
@@ -389,7 +538,7 @@ class FlowEnsemble:
                 batch_size,
                 optimizer,
                 loss_fn,
-                sample_errs,
+                convolve_errs,
                 seed,
                 verbose,
             )
